// Trellis platform tools module

#include <signal.h>
#include <unistd.h>
#include <sys/types.h>

#include "trellis.h"
#include "seesaw.h"
#include "keypad.h"

#define TRELLIS_BASE 0x0E
#define TRELLIS_STATUS 0x00
#define TRELLIS_PIN 0x01
#define TRELLIS_BUF_LEN 0x03
#define TRELLIS_BUF 0x04
#define TRELLIS_SHOW 0x05
#define TRELLIS_EVENT_RISING_EDGE 0x11
#define TRELLIS_EVENT_FALLING_EDGE 0x09
#define DEFAULT_TRELLIS_BUFFER_SIZE_H_8 0x00
#define DEFAULT_TRELLIS_BUFFER_SIZE_L_8 0x30
#define DEFAULT_TRELLIS_OUTPUT_PIN 0x3
#define DEFAULT_TRELLIS_SIZE 16
#define DEFAULT_EVENT_NUMBER 16
#define O_SHAPE_SIZE 8
#define X_SHAPE_SIZE 8
#define CURRENT_TURN_SIZE 8
// blink delay in miliseconds
#define DEFAULT_BLINK_DELAY 300


const int O_SHAPE_INDECES[] = {1, 2, 4, 7, 8, 11, 13, 14};
// x shape is simply two diagonals
const int X_SHAPE_INDECES[] = {0, 3, 5, 6, 9, 10, 12, 15};

const int CURRENT_TURN_INDECES[] = {3, 7, 11, 12, 13, 14, 15};

pid_t blinkPID = -1;
color LEDS[16] = {OFF};
color lastChosenColor = OFF;

bool compareColors(color c1, color c2) {
    return c1.r == c2.r && c1.g == c2.g && c1.b == c2.b;
}

static int getSeeSawKey(val) {
    return val / 8 * 4 + val % 8;
}

static void setBufferSize() {
    ssWrite(4, TRELLIS_BASE, TRELLIS_BUF_LEN,
                DEFAULT_TRELLIS_BUFFER_SIZE_H_8, DEFAULT_TRELLIS_BUFFER_SIZE_L_8);
}

static void setOutputPin() {
    ssWrite(3, TRELLIS_BASE, TRELLIS_PIN, DEFAULT_TRELLIS_OUTPUT_PIN);
}

// index - 1, if index is 1-ordered
void setPixelColor(int index, color c) {
    ssWrite(7, TRELLIS_BASE, TRELLIS_BUF, 0x00, index * 3, c.g, c.r, c.b);
}

static void setColorForAllPixels(color c) {
    for (int i = 0; i < DEFAULT_TRELLIS_SIZE; i++) {
        setPixelColor(i, c);
    }
}

void pixelOn() {
    ssWrite(2, TRELLIS_BASE, TRELLIS_SHOW);
}

void pixelOff(int index) {
    setPixelColor(index, OFF);
    pixelOn();
}

void setColorForPixels(int* indeces, size_t len, color c) {
    for (int i = 0; i < len; i++) {
        setPixelColor(indeces[i], c);
    }
}

void allPixelsOff() {
    for (int i = 0; i < DEFAULT_TRELLIS_SIZE; i++) {
        setPixelColor(i, OFF);
    }
    pixelOn();
}

// Display diagonally 4 colors. Len should be <= 4
// startIndex + 5 * i - to display on diagonal
void displayColorsOnDiagonal(const color* colors, size_t len, int startIndex) {
    for (int i = 0; i < len; i++) {
        setPixelColor(startIndex + 5 * i, colors[i]);
    }
    pixelOn();
}

void displayInStepWithDelay(int step, int delay, int* indeces, size_t len, color c) {
    setColorForPixels(indeces, len, OFF);
    pixelOn();
    sleepMS(delay);

    for (int i = 0; i < len; i += step) {
        setPixelColor(indeces[i], c);
        setPixelColor(indeces[i+1], c);
        pixelOn();
        sleepMS(delay);
    }
}

void blinkAllPixels(int times, int delay, color clr) {
    while (times > 0) {
        printf("in the loop now\n");
        setColorForAllPixels(clr);
        pixelOn();
        sleepMS(delay);
        allPixelsOff();
        sleepMS(delay);

        times--;
    }
}

void fadingBlinkOnTwoIndeces(int delay, int delta, int* indeces1,
                                int* indeces2, int len1, int len2, color clr1, color clr2) {
    allPixelsOff();

    while (delay > 0) {
        setColorForPixels(indeces1, len1, clr1);
        setColorForPixels(indeces2, len2, clr2);
        pixelOn();
        sleepMS(delay);

        allPixelsOff();
        pixelOn();
        sleepMS(delay);

        // need to play with the values for epic fading
        delay -= delta;
        delta -= 7;
        if (delta < 0) {
            delta = 5;
        }
    }
}

void blinkAtIndeces(int times, int delay, int* indeces, size_t len, color c) {
    printf("in blink\n");
    setColorForPixels(indeces, len, OFF);
    pixelOn();

    while(times > 0) {
        setColorForPixels(indeces, len, c);
        pixelOn();
        sleepMS(delay);

        setColorForPixels(indeces, len, OFF);
        pixelOn();
        sleepMS(delay);

        times--;
    }
}

void blinkAtIndex(int times, int delay, int index, color c) {
    setPixelColor(index, OFF);
    pixelOn();

    while(times > 0) {
        setPixelColor(index, c);
        pixelOn();
        sleepMS(delay);

        setPixelColor(index, OFF);
        pixelOn();
        sleepMS(delay);

        times--;
    }
}

void blinkOShape(int times, color c) {
    blinkAtIndeces(times, DEFAULT_BLINK_DELAY, X_SHAPE_INDECES, X_SHAPE_SIZE, c);
}

void blinkXShape(int times, color c) {
   blinkAtIndeces(times, DEFAULT_BLINK_DELAY, O_SHAPE_INDECES, O_SHAPE_SIZE, c);
}

void startBlinkingIndefinitely(int delay, int index, color c) {
    blinkPID = fork();

    if (blinkPID == 0) {
        while (true) {
            printf("blinking indefinitely");
            blinkAtIndex(1, delay, index, c);
        }
    }
}

void stopBlinkingIndefinitely() {
    if (blinkPID != -1) {
        kill(blinkPID, SIGKILL);
        blinkPID = -1;
    }
}

static void removePreviousEvents(char* buf) {
    unsigned char events[DEFAULT_EVENT_NUMBER] = {0x00};
    int count = readEventCount();
    int bytesRead = readBoard(events, getMin(count + 2, DEFAULT_EVENT_NUMBER));

    while (count > 0) {
        bytesRead = readBoard(events, getMin(count + 2, DEFAULT_EVENT_NUMBER));
        count = readEventCount();
        printf("I'M IN THE LOOOP CLEANING YOHOO");
    }

    printf("left initial clearance");
}

static void enablePushEventForAll() {
    setEventForAll(TRELLIS_EVENT_RISING_EDGE, true);
}

static void disablePushEventForAll() {
    setEventForAll(TRELLIS_EVENT_RISING_EDGE, false);
}


int getNextPushedButtonIndex(int* exit_sig) {
    unsigned char events[DEFAULT_EVENT_NUMBER] = {0x00};
    // PROBABLY WILL HAVE TO MOVE CLEANING PHASE OUTSIDE
    // AND DO IT BEFORE TURN PIXEL IS BLINKING
    // TO AVOID POSSIBLE INFINITE CLEANING WHEN USER
    // SIMPLY KEEPS PRESSING BUTTON INFINITELY
    // but it seems to work super fast
    // so user can't really make it happen
    removePreviousEvents(events);


    while (true && *exit_sig != 1) {
        int count = readEventCount();
        if (count != 0) {
            // + 2 just in case (idk:D)
            // printf("reading board\n");
            int bytesRead = readBoard(events, getMin(count + 2, DEFAULT_EVENT_NUMBER));
            for (int i = 0; i < bytesRead; i++) {
                // Check trellis file in blinka library for reference
                unsigned eventNumber = getSeeSawKey((events[i] >> 2) & 0x3F);
                int edge = events[i] & 0x3;
                if (eventNumber < DEFAULT_TRELLIS_SIZE) {
                    printf("%d event\n", eventNumber);
                    return eventNumber;
                }

            }
        }
        sleepMS(5);
    }

    return -1;
}


void initializeTrellis(const char* bus, unsigned char deviceAddress) {
    initializeSeeSawDevice(bus, deviceAddress);
    setBufferSize();
    enablePushEventForAll();
    pixelOn();
}

void destroyTrellis() {
    destroySeeSawDevice();
}
